/*
    buzzerv2.c
    Copyright (C) 2014
    Authors: Arun I, Hari Prasad Gokul R 

    This file is part of Quiz Buzzer v2 firmware.

    Quiz Buzzer v2 firmware is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Quiz Buzzer v2 firmware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Quiz Buzzer v2 firmware.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
  * @file
  *
  * Main program of the buzzer v2 firmware
  */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "ponseq.h"
#include "pins.h"

/**
 * Duty cycle of the seven segment display pulse width modulation
 *
 * Value can range from a minimum of 0 to a maximum of 255
 *
 * Duty cycle = PWM_DUTY_CYCLE / 256
 *
 * 0 corresponds to 0% duty cycle, 128 corresponds to 50% duty cycle, and 255 corresponds to 99.6% duty cycle
 */
#define PWM_DUTY_CYCLE 204

/**
 * Number of samples pin should be held low to be detected as low
 */
#define LOW_NO_OF_SAMPLES 10
/**
 * Time interval (in microseconds) between two successive samples of low detection
 */
#define LOW_SAMPLING_PERIOD 100

static void init(void);
static void display_ponsequence(void);

static uint8_t poll_for_team(void);
static void poll_for_masterreset(void);
static uint8_t heldlow(volatile uint8_t* pinreg, uint8_t pinno);
static uint8_t bcd2sevensegment(uint8_t bcd);

static void start_pwm(void);
static void stop_pwm(void);

/**
 * Team number currently displayed on the seven segment display
 */
uint8_t display_state = 0x00;

int main(void)
{
	init();
	display_ponsequence();
	while (1) {
		display_state = bcd2sevensegment(poll_for_team());	/// Poll for team
		DISPLAY_PORT = display_state;	/// Write team number to display
		DISPLAY_PORT |= _BV(BUZZER);	/// Sound buzzer
		start_pwm();			/// Start PWM of display output

		poll_for_masterreset();		/// Poll for master reset
		stop_pwm();			/// Stop PWM of display output
		DISPLAY_PORT = 0x00;		/// Clear display and stop buzzer
	}
	return 0;
}

static void init()
{
	/**
	 * Initialization routine
	 */
	// Initialize ports
	DDR(SWITCHES_PORT) = 0x00;	/// Configure SWITCHES_PORT as input
	SWITCHES_PORT = 0xff;		/// Configure SWITCHES_PORT to use internal pullup

	DDR(DISPLAY_PORT) = 0xff;	/// Configure DISPLAY_PORT as output
	DISPLAY_PORT = 0x00;		/// Reset DISPLAY_PORT

	DDR(MASTER_RESET_PORT) = 0x00;	// Configure MASTER_RESET_PORT as input
	MASTER_RESET_PORT = 0xff;	// Configure MASTER_RESET_PORT to use internal pullup

	/// Initialize timer
	TCCR2 &= ~_BV(FOC2);			// Disable force output compare
	TCCR2 |= _BV(WGM21) | _BV(WGM20);	// Waveform generation mode set to fast PWM
	TCCR2 &= ~(_BV(COM21) | _BV(COM20));	// Normal port operation, OC2 disconnected
	TCCR2 &= ~(_BV(CS22) | _BV(CS21) | _BV(CS20));	// Timer stopped
	OCR2 = PWM_DUTY_CYCLE;			// Set output compare value
	TIMSK |= _BV(OCIE2);			// Output compare match interrupt enable
	TIMSK |= _BV(TOIE2);			// Timer overflow interrupt enable
}

static void display_ponsequence()
{
	/**
	 * Display the power on sequence as defined in ponseq.h
	 */
	uint8_t i;
	for (i=0; i<len_of_ponseq; i++) {
		DISPLAY_PORT = ponseq[i];	// Display character
		DISPLAY_PORT |= _BV(BUZZER);	// Sound buzzer
		_delay_ms(200);
		DISPLAY_PORT &= ~_BV(BUZZER);	// Stop buzzer
		_delay_ms(800);
	}
	DISPLAY_PORT = 0x00;			// Clear display
}

static uint8_t poll_for_team()
{
	/**
	 * Blocks until any of team switches are pressed
	 *
	 * @return Team no of the switch which was pressed
	 */
	uint8_t switches, i;
	while (1) {
		switches = PIN(SWITCHES_PORT);	// Read switches
		// If some switch is pressed, find out which switch, else continue
		if (switches != 0xff) {
			for (i=0; i<8; i++) {
				if (!(switches & 0x01))
					// Check if switch is held low for a certain amount of time
					if (heldlow(&PIN(SWITCHES_PORT), i))
						return(i + 1);
				switches >>= 1;
			}
		}
	}
}

static void poll_for_masterreset()
{
	/**
	 * Blocks until master reset switch is pressed
	 */
	uint8_t masterreset;
	while (1) {
		masterreset = PIN(MASTER_RESET_PORT);
		if (!(masterreset & _BV(MASTER_RESET)))
			// Check if switch is held low for a certain amount of time
			if (heldlow(&PIN(MASTER_RESET_PORT), MASTER_RESET))
				return;
	}
}

static uint8_t heldlow(volatile uint8_t* pinreg, uint8_t pinno)
{
	/**
	 * Repeatedly samples a pin, and detects if it is being held low
	 *
	 * Samples the pin LOW_NO_OF_SAMPLES times at a sampling interval
	 * of LOW_SAMPLING_PERIOD. If the pin is low in all those samples,
	 * it is detected as having been held low. The purpose of this function
	 * is to low pass filter switch presses, and thus minimize the effects of noise.
	 *
	 * @param pinreg Address of PIN register to sample
	 * @param pinno Pin number (0 - 7) to sample
	 *
	 * @return 1 if pin is successfully held low in all samples, 0 otherwise
	 */
	uint8_t i;
	for (i=LOW_NO_OF_SAMPLES; i>0; i--) {
		if (*pinreg & _BV(pinno)) return 0;
		_delay_us(LOW_SAMPLING_PERIOD);
	}
	return 1;
}

static uint8_t bcd2sevensegment(uint8_t bcd)
{
	/**
	 * Converts given bcd (decimal number) to equivalent seven segment code
	 *
	 * @param bcd Decimal number to be converted
	 *
	 * @return Seven segment code corresponding to the given decimal number
	 */
	static const uint8_t map[] PROGMEM = {0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07, 0x7f, 0x6f};
	return pgm_read_byte(&map[bcd]);
}

static void start_pwm()
{
	/**
	 * Start pulse width modulation of display outputs
	 */
	sei();			/// Enable interrupts
	TCCR2 |= _BV(CS21);	/// Start timer, prescaling factor = 8 (PWM frequency = 500 Hz approx, for CLK_IO = 1 MHz)
}

static void stop_pwm()
{
	/**
	 * Stop pulse width modulation of display outputs
	 */
	cli();						/// Disable interrupts
	TCCR2 &= ~(_BV(CS22) | _BV(CS21) | _BV(CS20));	/// Stop timer
}

ISR(TIMER2_COMP_vect)
{
	/**
	 * Interrupt service routine for PWM timer compare match
	 */
	DISPLAY_PORT &= ~display_state;	/// Reset display output
}

ISR(TIMER2_OVF_vect)
{
	/**
	 * Interrupt service routine for PWM timer overflow
	 */
	DISPLAY_PORT |= display_state;	/// Set display output
}
