/*
    ponseq.c
    Copyright (C) 2014
    Authors: Arun I, Hari Prasad Gokul R 

    This file is part of Quiz Buzzer v2 firmware.

    Quiz Buzzer v2 firmware is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Quiz Buzzer v2 firmware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Quiz Buzzer v2 firmware.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
  * @file
  *
  * Program that generates ponseq.h based on the power on sequence given by the user
  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/**
 * Header file to be written to
 */
#define PONSEQ_HEADER "ponseq.h"

static void writetoheader(char *ponstring);

static short int mapdigit(char digit);
static short int mapalpha(char alpha);

static void print7segchar(short int segcode);
static void strtoupper(char* str);

int main()
{
	char* ponstring;

	/// Read the required power on sequence string
	printf("Enter the required power on sequence (no spaces): ");
	scanf("%ms", &ponstring);

	/// Write seven segment codes to header file
	writetoheader(ponstring);

	free(ponstring);
	return 0;
}

static void writetoheader(char *ponstring)
{
	/**
	 * Write seven segment codes representing the power on sequence to header file
	 *
	 * @param ponstring Power on sequence string
	 */
	int i, len = strlen(ponstring);
	short int segcode;
	FILE* header = fopen(PONSEQ_HEADER, "w");

	strtoupper(ponstring);

	/// Print general comments
	fprintf(header, "// Generated by ponseq\n\n");
	fprintf(header, "// Power on sequence is %s\n\n", ponstring);

	/// Print power on sequence
	fprintf(header, "/**\n * Sequence of seven segment codes corresponding to the power on sequence\n */\n");
	fprintf(header, "uint8_t ponseq[] = {");

	for (i=0; i<len; i++) {
		if isalpha(ponstring[i])
			segcode = mapalpha(ponstring[i]);
		else if isdigit(ponstring[i])
			segcode = mapdigit(ponstring[i]);
		else continue;
		fprintf(header, "0x%x, ", segcode);
		print7segchar(segcode);
	}
	fseek(header, -2, SEEK_CUR);	// Seek back to overwrite comma and space printed earlier
	fprintf(header, "};\n\n");

	/// Print length of power on sequence
	fprintf(header, "/**\n * Length (number of characters) of power on sequence\n */\n");
	fprintf(header, "uint8_t len_of_ponseq = %d;\n\n", i);

	fclose(header);
	printf("ponseq.h updated\n");
}

static short int mapdigit(char digit)
{
	/**
	 * Map decimal digit to seven segment display code
	 *
	 * @param digit Decimal digit character to be converted to seven segment display code
	 *
	 * @return Seven segment display code corresponding to the given decimal digit character
	 */
	short int map[] = {0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07, 0x7f, 0x6f};
	return map[digit - '0'];
}

static short int mapalpha(char alpha)
{
	/**
	 * Map alphabetic character to seven segment display code
	 *
	 * @param alpha Alphabetic character to be converted to seven segment display code
	 *
	 * @return Seven segment display code corresponding to the given alphabetic character
	 */
	short int map[] = {0x77, 0x7c, 0x39, 0x5e, 0x79, 0x71, 0x6f, 0x76, 0x30, 0x1e, 0x76, 0x38, 0x15, 
			   0x54, 0x3f, 0x73, 0x67, 0x50, 0x6d, 0x78, 0x3e, 0x1c, 0x2a, 0x76, 0x6e, 0x5b};
	return map[alpha - 'A'];
}

static void print7segchar(short int segcode)
{
	/**
	 * Print seven segment display representation to stdout
	 *
	 * @param segcode Seven segment display code
	 */
	printf(" ");
	if (segcode & (1<<0)) printf("_");
	printf("\n");
	if (segcode & (1<<5)) printf("|");
	else printf(" ");
	if (segcode & (1<<6)) printf("_");
	else printf(" ");
	if (segcode & (1<<1)) printf("|");
	printf("\n");
	if (segcode & (1<<4)) printf("|");
	else printf(" ");
	if (segcode & (1<<3)) printf("_");
	else printf(" ");
	if (segcode & (1<<2)) printf("|");
	printf("\n");
}

static void strtoupper(char* str)
{
	int i, len = strlen(str);
	for (i=0; i<len; i++)
		str[i] = toupper(str[i]);
}
